<?php
 /**
  * Copyright © Firebear Studio, Inc. All rights reserved.
  * See COPYING.txt for license details.
  */
namespace Firebear\PlatformSap\Model\Import\Source;

use Magento\Framework\Api\ExtensionAttributesFactory;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DB\Adapter\AdapterInterface;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Customer\Api\GroupRepositoryInterface;
use Magento\Customer\Api\Data\GroupInterface;
use Magento\Customer\Api\Data\GroupInterfaceFactory;
use Magento\Framework\Serialize\SerializerInterface;
use Magento\Framework\App\CacheInterface;
use Firebear\PlatformSap\Model\Import\Source\Gateway\Customer as CustomerGateway;
use Magento\Framework\App\RequestInterface;
use Magento\Customer\Model\ResourceModel\AddressRepository;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\Api\FilterBuilder;
use Magento\Customer\Model\ResourceModel\CustomerRepository;
use Firebear\PlatformSap\Model\Import\Source\Gateway\CustomerGroup as CustomerGroupGateway;
use Magento\Directory\Model\ResourceModel\Region\CollectionFactory as RegionCollectionFactory;
use Magento\Directory\Helper\Data as DirectoryHelper;
use Magento\Store\Api\WebsiteRepositoryInterface;

/**
 * Sap customer source
 */
class Customer extends AbstractSource
{
    const DEFAULT_GROUP_ID = 1;
    const DEFAULT_WEBSITE = 'base';
    const DEFAULT_WEBSITE_ID = 1;
    const DEFAULT_REGION = 'CO';
    const DEFAULT_POSTCODE = '12345';
    const DEFAULT_COUNTRY = 'US';
    const DEFAULT_TELEPHONE = '123456789';
    const SAP_ENTITY_GROUP_ENDPOINT = 'CustomerGroups/List';
    const CACHE_TAG = 'config_scopes';

    /**
     * @var CustomerGateway
     */
    protected $gateway;

    /**
     * Config data
     *
     * @var array
     */
    protected $_data = [];

    /**
     * @var RequestInterface
     */
    protected $request;

    /**
     * @var AddressRepository
     */
    protected $addressRepository;

    /**
     * @var FilterBuilder
     */
    protected $filterBuilder;

    /**
     * @var SearchCriteriaBuilder
     */
    protected $searchCriteriaBuilder;

    /**
     * @var ExtensionAttributesFactory
     */
    protected $extensionAttributesFactory;

    /**
     * @var GroupInterfaceFactory
     */
    protected $groupDataFactory;

    /**
     * Json Serializer
     *
     * @var SerializerInterface
     */
    protected $serializer;

    /**
     * @var CacheInterface
     */
    protected $cache;

    /**
     * @var RegionCollectionFactory
     */
    protected $regionsFactory;

    /**
     * @var DirectoryHelper
     */
    protected $directoryHelper;

    /**
     * @var WebsiteRepositoryInterface
     */
    protected $websiteRepository;

    /**
     * @var array
     */
    protected $websiteData = [];

    /**
     * @var array
     */
    private $addressEntityIds;

    /**
     * @var CustomerRepository
     */
    private $customerRepository;

    /**
     * @var array
     */
    private $sapCustomerGroups = [];

    /**
     * @var array
     */
    private $magentoCustomerGroups;

    /**
     * @var CustomerGroupGateway
     */
    private $customerGroupGateway;

    /**
     * @var GroupRepositoryInterface
     */
    private $groupRepository;

    /**
     * DB connection.
     *
     * @var AdapterInterface
     */
    private $connection;

    /**
     * @var ResourceConnection
     */
    private $resourceConnection;

    /**
     * @var array
     */
    private $importedAddresses;

    /**
     * @var array
     */
    private $regions;

    /**
     * Customer constructor.
     * @param CustomerGateway $gateway
     * @param RequestInterface $request
     * @param AddressRepository $addressRepository
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param FilterBuilder $filterBuilder
     * @param CustomerRepository $customerRepository
     * @param CustomerGroupGateway $customerGroupGateway
     * @param GroupInterfaceFactory $groupDataFactory
     * @param GroupRepositoryInterface $groupRepository
     * @param ResourceConnection $resourceConnection
     * @param SerializerInterface $serializer
     * @param CacheInterface $cache
     * @param RegionCollectionFactory $regionsFactory
     * @param DirectoryHelper $directoryHelper
     * @param WebsiteRepositoryInterface $websiteRepository
     * @param null $data
     * @throws \Exception
     */
    public function __construct(
        CustomerGateway $gateway,
        RequestInterface $request,
        AddressRepository $addressRepository,
        SearchCriteriaBuilder $searchCriteriaBuilder,
        FilterBuilder $filterBuilder,
        CustomerRepository $customerRepository,
        CustomerGroupGateway $customerGroupGateway,
        GroupInterfaceFactory $groupDataFactory,
        GroupRepositoryInterface $groupRepository,
        ResourceConnection $resourceConnection,
        SerializerInterface $serializer,
        CacheInterface $cache,
        RegionCollectionFactory $regionsFactory,
        DirectoryHelper $directoryHelper,
        WebsiteRepositoryInterface $websiteRepository,
        $data = null
    ) {
        $this->gateway = $gateway;
        $this->request = $request;
        $this->addressRepository = $addressRepository;
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
        $this->filterBuilder = $filterBuilder;
        $this->customerRepository = $customerRepository;
        $this->customerGroupGateway = $customerGroupGateway;
        $this->groupDataFactory = $groupDataFactory;
        $this->groupRepository = $groupRepository;
        $this->resourceConnection = $resourceConnection;
        $this->connection = $resourceConnection->getConnection();
        $this->serializer = $serializer;
        $this->cache = $cache;
        $this->regionsFactory = $regionsFactory;
        $this->directoryHelper = $directoryHelper;
        $this->websiteRepository = $websiteRepository;
        $this->parseEntities($data);
    }

    /**
     * @param $data
     * @return array|bool
     * @throws \Exception
     */
    protected function parseEntities($data)
    {
        $this->initWebsiteData();
        $this->initCustomerGroups($data);
        $this->entities = parent::parseEntities($data);
        if (!empty($this->entities)) {
            foreach ($this->entities as $key => $entity) {

                if (!$this->validateData($entity, '_website')) {
                    $this->entities[$key]['_website'] = self::DEFAULT_WEBSITE;
                }

                if ($this->validateData($entity, 'E_Mail')) {
                    $email = $this->filterEmail($entity['E_Mail']);
                    $this->entities[$key]['email'] = $email;
                    $this->initAddressEntityIds($email, $this->entities[$key]['_website']);
                } else {
                    unset($this->entities[$key]);
                    continue;
                }

                $nameData = $this->prepareNameData($entity);
                $this->entities[$key]['firstname'] = $nameData['firstname'];
                $this->entities[$key]['lastname'] = $nameData['lastname'];

                if ($this->validateData($entity, 'CardCode')) {
                    $this->entities[$key]['sap_id'] = $entity['CardCode'];
                }

                if ($this->validateData($entity, 'Phone1')) {
                    $this->entities[$key]['telephone'] = $entity['Phone1'];
                } else {
                    $this->entities[$key]['telephone'] = self::DEFAULT_TELEPHONE;
                }

                if ($this->validateData($entity, 'GroupCode')
                    && (isset($this->magentoCustomerGroups[$entity['GroupCode']]))
                ) {
                    $this->entities[$key]['group_id'] = $this->magentoCustomerGroups[$entity['GroupCode']];
                } else {
                    $this->entities[$key]['group_id'] = self::DEFAULT_GROUP_ID;
                }

                if ($this->validateData($entity, 'LicTradNum')) {
                    $this->entities[$key]['taxvat'] = $entity['LicTradNum'];
                }

                if ($this->validateData($entity, 'Address')) {
                    $this->entities[$key]['_address_sap_id'] = $entity['Address'];
                    if (isset($this->addressEntityIds[$email][$entity['Address']])) {
                        $addressEntityId = $this->addressEntityIds[$email][$entity['Address']];

                        if (isset($this->importedAddresses[$addressEntityId])) {
                            unset($this->entities[$key]);
                            continue;
                        }

                        $this->entities[$key]['_address_entity_id'] = $addressEntityId;
                        $this->importedAddresses[$addressEntityId] = true;
                    }
                    $this->entities[$key]['_address_firstname'] = $nameData['firstname'];
                    $this->entities[$key]['_address_lastname'] = $nameData['lastname'];
                    if ($this->validateData($entity, 'Street')) {
                        $this->entities[$key]['_address_street'] = $entity['Street'];
                    } else {
                        $this->entities[$key]['_address_street'] = 'Default';
                    }
                    if ($this->validateData($entity, 'City')) {
                        $this->entities[$key]['_address_city'] = $entity['City'];
                    } else {
                        $this->entities[$key]['_address_city'] = 'Default';
                    }
                    if ($this->validateData($entity, 'Country')) {
                        $this->entities[$key]['_address_country_id'] = $entity['Country'];
                    } else {
                        $this->entities[$key]['_address_country_id'] = self::DEFAULT_COUNTRY;
                    }
                    if ($this->validateData($entity, 'ZipCode')) {
                        $this->entities[$key]['_address_postcode'] = $entity['ZipCode'];
                    } elseif (!$this->directoryHelper->isZipCodeOptional($this->entities[$key]['_address_country_id'])) {
                        $this->entities[$key]['_address_postcode'] = self::DEFAULT_POSTCODE;
                    }
                    if ($this->validateData($entity, 'State')) {
                        if (!isset($this->regions[$this->entities[$key]['_address_country_id']])) {
                            $this->initRegions($this->entities[$key]['_address_country_id']);
                        }
                        if (isset($this->regions[$this->entities[$key]['_address_country_id']][$entity['State']])) {
                            $this->entities[$key]['_address_region'] = $entity['State'];
                        } elseif (empty($this->regions[$this->entities[$key]['_address_country_id']])
                            && $this->validateData($entity, 'StateName')) {
                            $this->entities[$key]['_address_region'] = $entity['StateName'];
                        }
                    } elseif ($this->entities[$key]['_address_country_id'] == 'US') {
                        $this->entities[$key]['_address_region'] = self::DEFAULT_REGION;
                    }
                    if ($this->validateData($entity, 'DefaultShipping') && $entity['DefaultShipping']) {
                        $this->entities[$key]['_address_default_shipping_'] = true;
                    }
                    if ($this->validateData($entity, 'DefaultBilling') && $entity['DefaultBilling']) {
                        $this->entities[$key]['_address_default_billing_'] = true;
                    }

                    if ($this->validateData($entity, 'AddressVatNumber')) {
                        $this->entities[$key]['_address_vat_id'] = $entity['AddressVatNumber'];
                    }
                }
            }
        }
        $this->entities = array_values($this->entities);
        $this->_colNames = array_keys($this->entities[0] ?? []);
        return $this->entities;
    }

    /**
     * @param $countryId
     */
    protected function initRegions($countryId)
    {
        $regionCollection = $this->regionsFactory->create();
        $regions = $regionCollection->addCountryFilter($countryId);
        if (!empty($regions)) {
            foreach ($regions as $region) {
                $this->regions[$countryId][$region->getCode()] = true;
            }
        }
    }

    /**
     * @param $email
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    private function initAddressEntityIds($email, $website)
    {
        if (!isset($this->addressEntityIds[$email])) {
            $addresses = [];
            try {
                if (isset($this->websiteData[$website])) {
                    $websiteId = $this->websiteData[$website];
                } else {
                    $websiteId = self::DEFAULT_WEBSITE_ID;
                }
                $customer = $this->customerRepository->get($email, $websiteId);
                $customerId = $customer->getId();
            } catch (NoSuchEntityException $exception) {
                $customerId = false;
            }
            if ($customerId) {
                $filter = $this->filterBuilder->setField('parent_id')
                    ->setValue($customerId)
                    ->setConditionType('eq')
                    ->create();
                $addresses = (array)($this->addressRepository->getList(
                    $this->searchCriteriaBuilder->addFilters([$filter])->create()
                )->getItems());
            }

            if (count($addresses)) {
                foreach ($addresses as $address) {
                    $sapInternalId = $address->getCustomAttribute('sap_id');
                    if ($sapInternalId && !empty($sapInternalId->getValue())) {
                        $this->addressEntityIds[$email][$sapInternalId->getValue()] = $address->getId();
                    }
                }
            }
        }
    }

    /**
     * @param $entity
     * @return array
     */
    private function prepareNameData($entity)
    {
        $fullName = 'Default Name';
        if (isset($entity['CardName'])) {
            $fullName = preg_replace('!\s+!', ' ', $entity['CardName']);
        }

        $fullName = explode(' ', $fullName);
        $firstname = $fullName[0];
        if (count($fullName) > 1) {
            $lastname = $fullName[1];
        } else {
            $lastname = $fullName[0];
        }

        return [
            'firstname' => $firstname,
            'lastname' => $lastname
        ];
    }

    /**
     *  Website data initialization
     */
    private function initWebsiteData()
    {
        if (empty($this->websiteData)) {
            foreach ($this->websiteRepository->getList() as $website) {
                $this->websiteData[$website->getCode()] = $website->getId();
            }
        }
    }

    /**
     * @param $data
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    private function initCustomerGroups($data)
    {
        if (empty($data['get_only_first_page'])) {
            $page = $this->request->getParam('offset', null);

            if (empty($page) || $page == 1) {
                $data['page'] = 1;
                $this->initSapCustomerGroups($data);
            } else {
                $sapCustomerGroups = $this->cache->load('sap_customer_groups');
                if ($sapCustomerGroups) {
                    $this->sapCustomerGroups = $this->serializer->unserialize($sapCustomerGroups);
                } else {
                    $this->initSapCustomerGroups($data);
                }
            }

            $this->initMagentoCustomerGroups();
            $this->createMagentoCustomerGroups();
        }
    }

    /**
     * @throws NoSuchEntityException
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\State\InvalidTransitionException
     */
    private function createMagentoCustomerGroups()
    {
        foreach ($this->sapCustomerGroups as $groupCode => $groupName) {
            if (!isset($this->magentoCustomerGroups[$groupCode])) {
                $group = $this->groupDataFactory->create();
                $group->setCode($groupName);
                $group = $this->groupRepository->save($group);
                $this->setSapInternalId($group, $groupCode);
                $this->magentoCustomerGroups[$groupCode] = $group->getId();
            }
        }
    }

    /**
     * @param $data
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    private function initSapCustomerGroups($data)
    {
        $sapCustomerGroups = '';
        while ($customerGroups = $this->customerGroupGateway->getSourceData($data)) {
            if (!empty($customerGroups)) {
                foreach ($customerGroups as $customerGroup) {
                    $this->sapCustomerGroups[$customerGroup['GroupCode']] = $customerGroup['GroupName'];
                }
                $data['page']++;
            } else {
                break;
            }
        }
        if (!empty($this->sapCustomerGroups)) {
            $sapCustomerGroups = $this->serializer->serialize($this->sapCustomerGroups);
        }
        $this->cache->save($sapCustomerGroups, 'sap_customer_groups', [self::CACHE_TAG]);
    }

    /**
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    private function initMagentoCustomerGroups()
    {
        $filter = $this->filterBuilder->setField('sap_id')
            ->setValue(null)
            ->setConditionType('neq')
            ->create();

        $magentoCustomerGroups = $this->groupRepository
            ->getList($this->searchCriteriaBuilder->addFilters([$filter])->create())->getItems();

        foreach ($magentoCustomerGroups as $magentoCustomerGroup) {
            $sapInternalId = $magentoCustomerGroup->getExtensionAttributes()->getSapId();
            $this->magentoCustomerGroups[$sapInternalId] = $magentoCustomerGroup->getId();
        }
    }

    /**
     * @param GroupInterface $group
     * @param $sapInternalId
     */
    private function setSapInternalId(GroupInterface $group, $sapInternalId)
    {
        $data = [
            'customer_group_id' => $group->getId(),
            'customer_group_code' => $group->getCode(),
            'tax_class_id' => $group->getTaxClassId(),
            'sap_id' => $sapInternalId
        ];

        $tableName = $this->resourceConnection->getTableName('customer_group');
        $this->connection->insertOnDuplicate($tableName, $data);
    }

    private function filterEmail($emailStr)
    {
        $email = filter_var($emailStr, FILTER_VALIDATE_EMAIL);

        if (!$email) {
            $pattern = '/[a-z][_a-z0-9-]+(\.[_a-z0-9-]+)*@([a-z0-9-]{2,})+(\.[a-z0-9-]{2,})*$/';
            preg_match($pattern, $emailStr, $matches);

            $email = isset($matches[0]) ? $matches[0] : $emailStr;
        }

        return $email;
    }
}
